package simple

import (
	"regexp"
	"fmt"
	"path/filepath"
	"strings"
	"os"
	"bytes"
	"bufio"
	"compress/gzip"
	"encoding/hex"
	"io"
	"io/ioutil"
	"crypto/md5"
	"crypto/sha1"
	"crypto/sha256"
	"github.com/blang/semver"
)

var (
	variableRegexp = regexp.MustCompile("(.*?):\\s*(.*)")
)

type PackageFile struct {
	Path string
	Name string
	Size int64
	MD5Hash string
	SHA1Hash string
	SHA256Hash string
	ControlData string
	Info *Package
}

type Package struct {
	Package string
	Version string
	Architecture string
}

func parsePackageData(ctlData string) *Package {
	res := &Package{}

	for _, match := range variableRegexp.FindAllStringSubmatch(ctlData, -1) {
		switch match[1] {
		case "Package":
			res.Package = match[2]
		case "Architecture":
			res.Architecture = match[2]
		case "Version":
			res.Version = match[2]
		}
	}

	return res
}

type Distro struct {
	Name string
	Architectures map[string]map[string]*PackageFile
}

var (
	distros map[string]*Distro = make(map[string]*Distro)
)

func buildPackageList(config Conf, distro, arch string) (map[string]*PackageFile, error) {
	m := make(map[string]*PackageFile)

	if err := scanRecursive(config.PoolPath(distro, arch), m); err != nil {
		return nil, err
	}

	return m, nil
}

func scanRecursive(path string, m map[string]*PackageFile) error {
	dirList, err := ioutil.ReadDir(path)

	if err != nil {
		return err
	}

	for _, file := range dirList {
		if file.IsDir() {
			if err := scanRecursive(filepath.Join(path, file.Name()), m); err != nil {
				return err
			}
			continue
		}

		if !strings.HasSuffix(file.Name(), "deb") {
			continue
		}

		p, err := newPackageFile(path, file.Name())

		if err != nil {
			return err
		}

		if old, exists := m[p.Info.Package]; exists {
			v1, err := semver.Parse(old.Info.Version)
			v2, err2 := semver.Parse(p.Info.Version)

			if err == nil && err2 == nil && v1.Compare(v2) > 0 {
				// Use old version
				continue
			}
		}

		m[p.Info.Package] = p
	}

	return nil
}

func newPackageFile(path, name string) (*PackageFile, error) {
	p := &PackageFile{
		Name: name,
		Path: filepath.Join(path, name),
	}

	debPath := filepath.Join(path, name)

	var err error

	p.ControlData, err = inspectPackage(debPath)

	if err != nil {
		return nil, err
	}

	p.Info = parsePackageData(p.ControlData)

	if stat, err := os.Stat(debPath); err == nil {
		p.Size = stat.Size()
	}

	f, err := os.Open(debPath)

	if err != nil {
		return nil, err
	}

	var (
		md5hash    = md5.New()
		sha1hash   = sha1.New()
		sha256hash = sha256.New()
	)

	w := io.MultiWriter(md5hash, sha1hash, sha256hash)

	if _, err := io.Copy(w, f); err != nil {
		return nil, err
	}

	p.MD5Hash = hex.EncodeToString(md5hash.Sum(nil))
	p.SHA1Hash = hex.EncodeToString(sha1hash.Sum(nil))
	p.SHA256Hash = hex.EncodeToString(sha256hash.Sum(nil))

	return p, nil
}

func createPackagesCached(config Conf, distro, arch string, packages map[string]*PackageFile) error {
	stdFile, err := os.Create(filepath.Join(config.ArchPath(distro, arch), "Packages"))
	if err != nil {
		return fmt.Errorf("failed to create packages: %s", err)
	}
	defer stdFile.Close()

	gzipFile, err := os.Create(filepath.Join(config.ArchPath(distro, arch), "Packages.gz"))
	if err != nil {
		return fmt.Errorf("failed to create packages.gz: %s", err)
	}
	defer gzipFile.Close()

	gzWriter := gzip.NewWriter(gzipFile)
	defer gzWriter.Close()

	stdOut := bufio.NewWriter(stdFile)

	// loop through each directory
	// run inspectPackage

	for _, p := range packages {
		var packBuf bytes.Buffer

		packBuf.WriteString(p.ControlData)

		dir := filepath.Join("pool/main", distro, arch, p.Info.Package[0:1], p.Info.Package, p.Name)

		fmt.Fprintf(&packBuf, "Filename: %s\n", strings.Replace(dir, "\\", "/", -1))
		fmt.Fprintf(&packBuf, "Size: %d\n", p.Size)

		fmt.Fprintf(&packBuf, "MD5sum: %s\n", p.MD5Hash)
		fmt.Fprintf(&packBuf, "SHA1: %s\n", p.SHA1Hash)
		fmt.Fprintf(&packBuf, "SHA256: %s\n", p.SHA256Hash)
		packBuf.WriteString("\n\n")

		stdOut.Write(packBuf.Bytes())
		gzWriter.Write(packBuf.Bytes())
	}

	stdOut.Flush()

	gzWriter.Flush()

	return nil
}